import asyncio
from datetime import timedelta
import os
import tempfile
from typing import List, Optional

import aiohttp
from rich.console import Console
from rich.progress import track

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    LogDownloadConfig,
    LogDownloadRequest,
    LogDownloadResult,
    LogFileChunk,
    LogFilter,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cluster import Cluster
from anyscale.sdk.anyscale_client.models.session_state import SessionState
from anyscale.utils.logs_utils import LogGroup


# Page size for listing bucket
DEFAULT_PAGE_SIZE = 1000

# Timeout for API Requests
DEFAULT_TIMEOUT = 30

# Timeout for Downloading Files
DEFAULT_READ_TIMEOUT = 30

DEFAULT_TTL = 14400
DEFAULT_PARALLELISM = 10


class LogsController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.console = Console()

    def render_logs(
        self, log_group: LogGroup, parallelism: int, read_timeout: timedelta, tail: int
    ):
        self._download_or_stdout(
            log_group=log_group,
            parallelism=parallelism,
            read_timeout=read_timeout,
            write_to_stdout=True,
            tail=tail,
            show_progress_bar=False,
        )

    def get_log_group(
        self,
        filter: LogFilter,  # noqa: A002
        page_size: Optional[int],
        ttl_seconds: Optional[int],
        timeout: timedelta,
    ) -> LogGroup:
        if filter.cluster_id:
            cluster: Cluster = self.anyscale_api_client.get_cluster(
                filter.cluster_id
            ).result
            if cluster.state == SessionState.RUNNING:
                self.log.warning(
                    "The latest 24 hours of logs are not guaranteed if the cluster is still running. "
                    "If so, please view logs via the Anyscale UI or Ray dashboard."
                )
        chunks = self._list_log_chunks(
            log_filter=filter,
            page_size=page_size,
            ttl_seconds=ttl_seconds,
            timeout=timeout,
        )
        log_group = self._group_log_chunk_list(chunks)
        return log_group

    def download_logs(
        self,
        # Provide filters
        filter: LogFilter,  # noqa: A002
        # List files config
        page_size: int = DEFAULT_PAGE_SIZE,
        ttl_seconds: int = DEFAULT_TTL,
        timeout: timedelta = timedelta(seconds=DEFAULT_TIMEOUT),
        read_timeout: timedelta = timedelta(seconds=DEFAULT_READ_TIMEOUT),
        # Download config
        parallelism: int = DEFAULT_PARALLELISM,
        download_dir: Optional[str] = None,
        write_to_stdout: bool = False,
    ) -> None:
        log_group = self.get_log_group(filter, page_size, ttl_seconds, timeout)
        self.console.log(
            f"Discovered {len(log_group.get_files())} log files across {len(log_group.get_chunks())} chunks."
        )

        self._download_or_stdout(
            download_dir=download_dir,
            read_timeout=read_timeout,
            parallelism=parallelism,
            log_group=log_group,
            show_progress_bar=True,
            write_to_stdout=write_to_stdout,
        )

    # TODO (shomilj): Refactor this method. It's too nested.
    def _download_or_stdout(
        self,
        log_group: LogGroup,
        parallelism: int,
        read_timeout: timedelta,
        download_dir: Optional[str] = None,
        write_to_stdout: bool = False,
        tail: int = -1,
        show_progress_bar: bool = False,
    ):
        if len(log_group.get_chunks()) == 0:
            return

        with tempfile.TemporaryDirectory() as tmp_dir:
            # Download all files to a temporary directory.
            asyncio.run(
                # TODO (shomilj): Add efficient tailing method here.
                self._download_files(
                    base_folder=tmp_dir,
                    log_chunks=log_group.get_chunks(),
                    parallelism=parallelism,
                    read_timeout=read_timeout,
                    show_progress_bar=show_progress_bar,
                )
            )

            for log_file in log_group.get_files():
                is_tail = tail > 0
                chunks = log_file.get_chunks(reverse=is_tail)
                if write_to_stdout:
                    # Write to standard out
                    lines_read = 0
                    for chunk in chunks:
                        with open(os.path.join(tmp_dir, chunk.chunk_name)) as source:
                            if tail > 0:
                                # Tail is enabled, so read log lines in reverse.
                                # TODO (shomilj): Make this more efficient (don't read everything into memory before reversing it)
                                # For now, this is fine (the chunks are <10 MB, so this isn't that inefficient).
                                for line in reversed(source.readlines()):
                                    print(line.strip())
                                    lines_read += 1
                                    if lines_read >= tail:
                                        return
                            else:
                                # Read log lines normally.
                                print(source.read())
                else:
                    # Write to destination files
                    real_path = os.path.join(
                        download_dir or "", log_file.get_target_path()
                    )
                    real_dir = os.path.dirname(real_path)
                    if not os.path.exists(real_dir):
                        os.makedirs(real_dir)

                    chunks_written = 0
                    with open(real_path, "w") as dest:
                        for chunk in chunks:
                            downloaded_chunk_path = os.path.join(
                                tmp_dir, chunk.chunk_name
                            )
                            if not os.path.exists(downloaded_chunk_path):
                                self.log.error(
                                    "Download failed for file: ", chunk.chunk_name
                                )
                                continue
                            with open(downloaded_chunk_path, errors="ignore") as source:
                                for line in source:
                                    dest.write(line)
                                dest.write("\n")
                            chunks_written += 1

                    if chunks_written == 0:
                        os.remove(real_path)

            if not write_to_stdout:
                if not download_dir:
                    download_dir = os.getcwd()
                sample_chunk = log_group.get_chunks()[0]
                cluster_id = sample_chunk.cluster_id
                download_dir = download_dir + f"/logs/{cluster_id}"
                self.console.log(
                    f"Download complete! Files have been downloaded to {download_dir}"
                )

    def _group_log_chunk_list(self, chunks: List[LogFileChunk]) -> LogGroup:
        # This has to happen locally because it happens after we retrieve all file metadata through the paginated
        # backend API for listing S3/GCS buckets.
        group = LogGroup()
        for chunk in chunks:
            group.insert_chunk(chunk=chunk)
        return group

    def _list_log_chunks(
        self,
        log_filter: LogFilter,
        page_size: Optional[int],
        ttl_seconds: Optional[int],
        timeout: timedelta,
    ) -> List[LogFileChunk]:
        next_page_token: Optional[str] = None
        all_log_chunks: List[LogFileChunk] = []

        with self.console.status("Scanning available logs...") as status:
            while True:
                request = LogDownloadRequest(
                    filter=log_filter,
                    config=LogDownloadConfig(
                        next_page_token=next_page_token,
                        page_size=page_size,
                        ttl_seconds=ttl_seconds,
                    ),
                )
                result: LogDownloadResult = self.api_client.get_log_files_api_v2_logs_get_log_files_post(
                    log_download_request=request, _request_timeout=timeout
                ).result
                all_log_chunks.extend(result.log_chunks)
                if status:
                    status.update(
                        f"Scanning available logs...discovered {len(all_log_chunks)} log file chunks."
                    )
                if (
                    result.next_page_token is None
                    or result.next_page_token == next_page_token
                ):
                    break
                next_page_token = result.next_page_token

        return all_log_chunks

    async def _download_file(
        self,
        sem: asyncio.Semaphore,
        pos: int,  # noqa: ARG002
        file_name: str,
        url: str,
        size: int,  # noqa: ARG002
        session: aiohttp.ClientSession,
        read_timeout: timedelta,
    ) -> None:
        async with sem:
            download_dir = os.path.dirname(file_name)
            if download_dir and not os.path.exists(download_dir):
                os.makedirs(download_dir)

            timeout = aiohttp.ClientTimeout(
                total=None, sock_connect=30, sock_read=read_timeout.seconds
            )
            async with session.get(url, timeout=timeout) as response:
                if response.status == 200:
                    with open(file_name, "wb") as fhand:
                        async for chunk in response.content.iter_chunked(1024):
                            fhand.write(chunk)
                else:
                    self.log.error(
                        f"Unable to download file {file_name}! response: [{response.status}, {await response.text()}]"
                    )

    async def _download_files(
        self,
        base_folder: Optional[str],
        log_chunks: List[LogFileChunk],
        parallelism: int,
        read_timeout: timedelta,
        show_progress_bar: bool = False,
    ) -> List[str]:
        sem = asyncio.Semaphore(parallelism)
        downloads = []
        connector = aiohttp.TCPConnector(limit_per_host=parallelism)
        paths = []
        async with aiohttp.ClientSession(connector=connector) as session:
            for pos, log_chunk in enumerate(log_chunks):
                path = os.path.join(base_folder or "", log_chunk.chunk_name.lstrip("/"))
                paths.append(path)
                downloads.append(
                    asyncio.create_task(
                        self._download_file(
                            sem,
                            pos,
                            path,
                            log_chunk.chunk_url,
                            log_chunk.size,
                            session,
                            read_timeout,
                        )
                    )
                )

            if show_progress_bar:
                for task in track(
                    asyncio.as_completed(downloads),
                    description="Downloading...",
                    total=len(downloads),
                    transient=True,
                ):
                    await task
            else:
                await asyncio.gather(*downloads)

        return paths



from datetime import datetime
import json
import random
import re
import string
import sys
from typing import Callable, Dict, List, Optional
from unittest.mock import Mock, patch

import boto3
from botocore.exceptions import ClientError
from moto import mock_iam
import pytest

from anyscale.cloud_resource import (
    aws_subnet_has_enough_capacity,
    verify_aws_cloudformation_stack,
    verify_aws_efs,
    verify_aws_iam_roles,
    verify_aws_s3,
    verify_aws_security_groups,
    verify_aws_subnets,
    verify_aws_vpc,
)
from anyscale.conf import ANYSCALE_IAM_ROLE_NAME
from anyscale.shared_anyscale_utils.aws import AwsRoleArn
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST
from frontend.cli.anyscale.aws_iam_policies import (
    AMAZON_ECR_READONLY_ACCESS_POLICY_ARN,
    AMAZON_S3_FULL_ACCESS_POLICY_ARN,
    ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN,
    ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
    get_anyscale_aws_iam_assume_role_policy,
)
from frontend.cli.anyscale.cli_logger import BlockLogger
from frontend.cli.anyscale.client.openapi_client.models.create_cloud_resource import (
    CreateCloudResource,
)


DEFAULT_RAY_IAM_ROLE = "ray-autoscaler-v1"


def _create_roles(anyscale_account: str):
    iam = boto3.client("iam")
    suffix = "".join(random.choices(string.ascii_lowercase, k=10))
    control_plane_role = iam.create_role(
        RoleName=ANYSCALE_IAM_ROLE_NAME + suffix,
        AssumeRolePolicyDocument=json.dumps(
            get_anyscale_aws_iam_assume_role_policy(
                anyscale_aws_account=anyscale_account
            )
        ),
    )["Role"]["Arn"]

    # Create & Configure the Dataplane Role
    dataplane_role_name = DEFAULT_RAY_IAM_ROLE + suffix
    dataplane_role = iam.create_role(
        RoleName=dataplane_role_name, AssumeRolePolicyDocument="{}"
    )["Role"]["Arn"]

    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN
    )
    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN
    )
    return [control_plane_role, dataplane_role]


def generate_cloud_resource_mock_aws() -> CreateCloudResource:
    return CreateCloudResource(
        aws_vpc_id="fake_aws_vpc_id",
        aws_subnet_ids=["fake_aws_subnet_id_0", "fake_aws_subnet_id_1"],
        aws_iam_role_arns=[
            "arn:aws:iam::123:role/mock_anyscale_role",
            "arn:aws:iam::123:role/mock_dataplane_role",
        ],
        aws_security_groups=["fake_aws_security_group_0"],
        aws_s3_id="fake_aws_s3_id",
        aws_efs_id="fake_aws_efs_id",
        aws_cloudformation_stack_id="fake_aws_cloudformation_stack_id",
    )


def _attach_iam_roles_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    steady_state = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=steady_state)

    initial_run = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=initial_run)


def _attach_inline_iam_policy_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )
    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )


def _create_instance_profile_for_role(
    role_arn: str, instance_profile_name: Optional[str]
) -> None:
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()
    if instance_profile_name is None:
        instance_profile_name = role_name

    iam = boto3.client("iam")
    iam.create_instance_profile(InstanceProfileName=instance_profile_name)
    iam.add_role_to_instance_profile(
        RoleName=role_name, InstanceProfileName=instance_profile_name,
    )


def generate_aws_security_groups() -> List[Mock]:
    ip_permission_443 = {"FromPort": 443}
    ip_permission_22 = {"FromPort": 22}
    inbound_ip_permissions = [
        ip_permission_443,
        ip_permission_22,
        {
            "IpProtocol": "-1",
            "UserIdGroupPairs": [{"GroupId": "fake_aws_security_group_0"}],
        },
    ]
    anyscale_security_group = Mock(ip_permissions=inbound_ip_permissions)
    return anyscale_security_group


def generate_aws_subnets() -> List[Mock]:
    return [
        Mock(
            id="mock_id_0",
            cidr_block="0.0.0.0/18",
            vpc_id="fake_aws_vpc_id",
            availability_zone="us-west-2a",
        ),
        Mock(
            id="mock_id_1",
            cidr_block="0.0.0.0/19",
            vpc_id="fake_aws_vpc_id",
            availability_zone="us-west-2c",
        ),
        Mock(
            id="mock_id_2",
            cidr_block="1.2.3.4/20",
            vpc_id="fake_aws_vpc_id",
            availability_zone="us-west-2b",
        ),
    ]


def generate_network_interfaces_mock(cloud_resource: CreateCloudResource) -> List[Mock]:
    return [
        Mock(
            subnet_id=subnet_id,
            groups=[
                {"GroupId": security_group_id}
                for security_group_id in cloud_resource.aws_security_groups
            ],
        )
        for subnet_id in cloud_resource.aws_subnet_ids
    ]


@pytest.mark.parametrize(
    ("vpc_exists", "vpc_cidr_block", "expected_result", "expected_log_message"),
    [
        pytest.param(False, "0.0.0.0/0", False, r"does not exist."),
        # Happy sizes
        pytest.param(True, "0.0.1.0/0", True, r"verification succeeded."),
        pytest.param(True, "0.0.2.0/20", True, r"verification succeeded."),
        # Warn sizes
        pytest.param(
            True, "0.0.3.0/21", True, r"but this vpc only supports up to \d+ addresses",
        ),
        pytest.param(
            True, "0.0.4.0/24", True, r"but this vpc only supports up to \d+ addresses",
        ),
        # Error sizes
        pytest.param(
            True,
            "0.0.4.0/25",
            False,
            r"Please reach out to support if this is an issue!",
        ),
    ],
)
def test_verify_aws_vpc(
    capsys,
    vpc_exists: bool,
    vpc_cidr_block: str,
    expected_result: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    vpc_mock = Mock(cidr_block=vpc_cidr_block) if vpc_exists else Mock()
    if not vpc_exists:
        vpc_mock.load = Mock(
            side_effect=ClientError(
                error_response={"Error": {"Code": "InvalidVpcID.NotFound"}},
                operation_name="DescribeVpcs",
            )
        )
    ec2_mock = Mock(Vpc=Mock(return_value=vpc_mock))
    boto3_session_mock = Mock(resource=Mock(return_value=ec2_mock))

    result = verify_aws_vpc(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3_session_mock,
        logger=BlockLogger(),
    )
    assert result == expected_result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


def test_verify_aws_subnets(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    subnets_mock = generate_aws_subnets()
    with patch.multiple(
        "anyscale.cloud_resource",
        _get_subnets_from_subnet_ids=Mock(return_value=subnets_mock),
    ):
        result, _ = verify_aws_subnets(
            cloud_resource=cloud_resource_mock,
            region="fake_region",
            is_private_network=False,
            logger=BlockLogger(),
        )
        assert result

        _, stderr = capsys.readouterr()
        assert re.search(r"verification succeeded.", stderr)


def test_verify_aws_subnets_insufficient_subnets(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cloud_resource_mock.aws_subnet_ids = cloud_resource_mock.aws_subnet_ids[:1]
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None
    result, returned_cloud_resource = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region="fake_region",
        is_private_network=False,
        logger=BlockLogger(),
    )
    assert result is False
    assert returned_cloud_resource is None
    _, stderr = capsys.readouterr()
    assert re.search(r"Need at least 2 subnets for a cloud.", stderr)


def test_verify_aws_subnets_no_subnets():
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cloud_resource_mock.aws_subnet_ids = None
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None
    result, returned_cloud_resource = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region="fake_region",
        is_private_network=False,
        logger=BlockLogger(),
    )
    assert result is False
    assert returned_cloud_resource is None


@pytest.mark.parametrize(
    ("subnet_cidr", "has_capacity", "expected_log_message"),
    [
        # Happy sizes
        ("0.0.0.0/0", True, None),
        ("0.0.1.0/24", True, None),
        # Warn sizes
        ("0.0.2.0/25", True, r"but this subnet only supports up to \d+ addresses"),
        ("0.0.3.0/28", True, r"but this subnet only supports up to \d+ addresses"),
        # Error sizes
        ("0.0.3.0/29", False, r"Please reach out to support if this is an issue!"),
    ],
)
def test_aws_subnet_has_enough_capacity(
    subnet_cidr, has_capacity, expected_log_message, capsys
):
    subnet = Mock()
    subnet.id = "vpc-fake_id"
    subnet.cidr_block = subnet_cidr

    assert aws_subnet_has_enough_capacity(subnet, BlockLogger()) == has_capacity

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@mock_iam
@pytest.mark.parametrize(
    ("iam_roles_mock_fn", "expect_warning"),
    [
        pytest.param(lambda _: None, True, id="No Policies"),
        pytest.param(_attach_iam_roles_to_role, False, id="Separate Policies"),
        pytest.param(_attach_inline_iam_policy_to_role, False, id="Inline Policies"),
    ],
)
def test_verify_aws_iam_roles(
    iam_roles_mock_fn: Callable[[str], None], expect_warning: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    _create_instance_profile_for_role(dp_role_arn, None)
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    iam_roles_mock_fn(cp_role_arn)
    mock_logger = Mock()
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    if expect_warning:
        assert len(mock_logger.warning.mock_calls) == 1
    else:
        mock_logger.warning.assert_not_called()

    assert len(mock_logger.info.mock_calls) == 2


@mock_iam
def test_verify_aws_iam_roles_instance_profiles():
    """Tests that the dataplane role only passes validation if it has
    an InstanceProfile with the same name as the Role."""
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    mock_logger = Mock()

    # No instance profile for role
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    # Create an instance profile for the role, but with the wrong name
    _create_instance_profile_for_role(dp_role_arn, "wrong_name")
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    # Create a correct insance profile with the role!
    _create_instance_profile_for_role(dp_role_arn, None)
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )


def test_verify_aws_security_groups():
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    security_group_mock = generate_aws_security_groups()
    ec2_mock = Mock(SecurityGroup=Mock(return_value=security_group_mock))
    boto3_session_mock = Mock(resource=Mock(return_value=ec2_mock))

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3_session_mock,
        logger=Mock(),
    )
    assert result


@pytest.mark.parametrize(
    ("s3_exists", "expected_result"),
    [
        pytest.param(False, False, id="S3-Does-Not-Exist"),
        pytest.param(True, True, id="S3-Exists"),
    ],
)
@pytest.mark.parametrize(
    "cors_rules,expected_warning_call_count",
    [
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": [ANYSCALE_HOST],
                    "ExposeHeaders": [],
                }
            ],
            0,
            id="CorrectCORS",
        ),
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": [],
                    "ExposeHeaders": [],
                }
            ],
            1,
            id="ImproperCORS",
        ),
        pytest.param([{}], 1, id="NoCORS"),
    ],
)
def test_verify_aws_s3(
    s3_exists: bool,
    expected_result: bool,
    cors_rules: List[Dict[str, List[str]]],
    expected_warning_call_count: int,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    s3_mock = Mock()
    s3_bucket_mock = (
        Mock(Cors=Mock(return_value=Mock(cors_rules=cors_rules)),)
        if s3_exists
        else Mock()
    )
    s3_bucket_mock.creation_date = datetime.now() if s3_exists else None
    s3_mock = Mock(Bucket=Mock(return_value=s3_bucket_mock))
    boto3_session_mock = Mock(resource=Mock(return_value=s3_mock))
    mock_logger = Mock()

    with patch.multiple(
        "anyscale.cloud_resource", verify_s3_access=Mock(return_value=True),
    ):
        result = verify_aws_s3(
            cloud_resource=cloud_resource_mock,
            boto3_session=boto3_session_mock,
            logger=mock_logger,
        )

    if s3_exists:
        assert mock_logger.warning.call_count == expected_warning_call_count
    else:
        assert mock_logger.error.call_count == 1
        assert mock_logger.warning.call_count == 0

    assert result == expected_result


@pytest.mark.parametrize(
    ("efs_exists", "expected_result"),
    [pytest.param(False, False), pytest.param(True, True)],
)
def test_verify_aws_efs(efs_exists: bool, expected_result: bool):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    efs_client_mock = Mock(
        describe_file_systems=Mock(
            return_value={"FileSystems": [{"k": "v"}]} if efs_exists else {}
        ),
        describe_mount_targets=Mock(return_value={"MountTargets": Mock()}),
    )
    boto3_session_mock = Mock(client=Mock(return_value=efs_client_mock),)
    network_interfaces_mock = generate_network_interfaces_mock(cloud_resource_mock)

    with patch.multiple(
        "anyscale.cloud_resource",
        _get_network_interfaces_from_mount_targets=Mock(
            return_value=network_interfaces_mock
        ),
    ):
        result = verify_aws_efs(
            cloud_resource=cloud_resource_mock,
            boto3_session=boto3_session_mock,
            logger=Mock(),
        )
        assert result == expected_result


@pytest.mark.parametrize(
    ("cloudformation_stack_exists", "expected_result"),
    [pytest.param(False, False), pytest.param(True, True)],
)
def test_verify_aws_cloudformation_stack(
    cloudformation_stack_exists: bool, expected_result: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cloudformation_stack_mock = Mock() if cloudformation_stack_exists else None
    cloudformation_mock = Mock(Stack=Mock(return_value=cloudformation_stack_mock))
    boto3_session_mock = Mock(resource=Mock(return_value=cloudformation_mock))

    result = verify_aws_cloudformation_stack(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3_session_mock,
        logger=Mock(),
    )
    assert result == expected_result



import inspect
import logging
import os
from sys import path
from typing import Any, Dict, List

from anyscale import snapshot_util, version


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(os.environ.get("ANYSCALE_LOGLEVEL", "WARN"))

anyscale_dir = os.path.dirname(os.path.abspath(__file__))
path.append(os.path.join(anyscale_dir, "client"))
path.append(os.path.join(anyscale_dir, "sdk"))

from anyscale import integrations
from anyscale.cluster import get_job_submission_client_cluster_info  # noqa: E402
from anyscale.connect import ClientBuilder  # noqa: E402
from anyscale.sdk.anyscale_client.sdk import AnyscaleSDK  # noqa: E402


# Note: indentation here matches that of connect.py::ClientBuilder.
BUILDER_HELP_FOOTER = """
        See ``anyscale.ClientBuilder`` for full documentation of
        this experimental feature."""

# Auto-add all Anyscale connect builder functions to the top-level.
for attr, _ in inspect.getmembers(ClientBuilder, inspect.isfunction):
    if attr.startswith("_"):
        continue

    def _new_builder(attr: str) -> Any:
        target = getattr(ClientBuilder, attr)

        def new_session_builder(*a: List[Any], **kw: Dict[str, Any]) -> Any:
            builder = ClientBuilder()
            return target(builder, *a, **kw)

        new_session_builder.__name__ = attr
        new_session_builder.__doc__ = target.__doc__ + BUILDER_HELP_FOOTER
        new_session_builder.__signature__ = inspect.signature(target)  # type: ignore

        return new_session_builder

    globals()[attr] = _new_builder(attr)

__version__ = version.__version__

ANYSCALE_ENV = os.environ.copy()

import anyscale.job



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Dict, Generic, List, TypeVar

import humanize
import tabulate

from anyscale.client.openapi_client.models.decorated_schedule import DecoratedSchedule


TR = TypeVar("TR")


class AbstractTable(ABC, Generic[TR]):
    """An Abstract table class """

    entity_name = "object"

    def __init__(self, rows: List[TR]):
        self.rows = rows

    @abstractmethod
    def render_row(self, row: TR) -> Dict[str, str]:
        """Render one row of the table

        Args:
            row (TR): A row object.

        Returns:
            Dict[str, str]: A map from column headers to row values. Ordering is preserved. The order of the headers
                in the dict is the order that will be shown in the table. This function should always return a dict with the same keys,
                in the same order.
        """
        ...

    def __str__(self):
        if len(self.rows) == 0:
            return f"No {self.entity_name}s found."
        rendered_rows = [self.render_row(row) for row in self.rows]
        headers = list(rendered_rows[0].keys())
        rows = [row.values() for row in rendered_rows]
        return tabulate.tabulate(rows, headers=headers, tablefmt="psql")


class SchedulesTable(AbstractTable[DecoratedSchedule]):
    entity_name = "Schedule"

    def render_row(self, decorated_schedule: DecoratedSchedule):
        last_exec_list = decorated_schedule.last_executions
        most_recent_exec_id = last_exec_list[0].id if len(last_exec_list) > 0 else ""

        return {
            "ID": decorated_schedule.id,
            "NAME": decorated_schedule.name,
            "DESCRIPTION": decorated_schedule.description,
            "PROJECT": decorated_schedule.project.name,
            "CRON": decorated_schedule.schedule.cron_expression,
            "NEXT TRIGGER": humanize.naturaltime(
                datetime.now(timezone.utc) - decorated_schedule.next_trigger_at
            )
            if decorated_schedule.next_trigger_at
            else "Schedule is paused",
            "TIMEZONE": decorated_schedule.schedule.timezone,
            "CREATOR": decorated_schedule.creator.email,
            "LATEST EXECUTION ID": most_recent_exec_id,
        }



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class JobsQuery(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'TextQuery',
        'runtime_environment_id': 'str',
        'cluster_id': 'str',
        'creator_id': 'str',
        'ray_job_id': 'str',
        'project_id': 'str',
        'include_child_jobs': 'bool',
        'ha_job_id': 'str',
        'show_ray_client_runs_only': 'bool',
        'paging': 'PageQuery',
        'state_filter': 'list[BaseJobStatus]',
        'type_filter': 'list[JobRunType]',
        'sort_by_clauses': 'list[SortByClauseJobsSortField]'
    }

    attribute_map = {
        'name': 'name',
        'runtime_environment_id': 'runtime_environment_id',
        'cluster_id': 'cluster_id',
        'creator_id': 'creator_id',
        'ray_job_id': 'ray_job_id',
        'project_id': 'project_id',
        'include_child_jobs': 'include_child_jobs',
        'ha_job_id': 'ha_job_id',
        'show_ray_client_runs_only': 'show_ray_client_runs_only',
        'paging': 'paging',
        'state_filter': 'state_filter',
        'type_filter': 'type_filter',
        'sort_by_clauses': 'sort_by_clauses'
    }

    def __init__(self, name=None, runtime_environment_id=None, cluster_id=None, creator_id=None, ray_job_id=None, project_id=None, include_child_jobs=False, ha_job_id=None, show_ray_client_runs_only=True, paging=None, state_filter=[], type_filter=[], sort_by_clauses=[{"sort_field":"CREATED_AT","sort_order":"DESC"},{"sort_field":"NAME","sort_order":"ASC"},{"sort_field":"ID","sort_order":"ASC"}], local_vars_configuration=None):  # noqa: E501
        """JobsQuery - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._runtime_environment_id = None
        self._cluster_id = None
        self._creator_id = None
        self._ray_job_id = None
        self._project_id = None
        self._include_child_jobs = None
        self._ha_job_id = None
        self._show_ray_client_runs_only = None
        self._paging = None
        self._state_filter = None
        self._type_filter = None
        self._sort_by_clauses = None
        self.discriminator = None

        if name is not None:
            self.name = name
        if runtime_environment_id is not None:
            self.runtime_environment_id = runtime_environment_id
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if creator_id is not None:
            self.creator_id = creator_id
        if ray_job_id is not None:
            self.ray_job_id = ray_job_id
        if project_id is not None:
            self.project_id = project_id
        if include_child_jobs is not None:
            self.include_child_jobs = include_child_jobs
        if ha_job_id is not None:
            self.ha_job_id = ha_job_id
        if show_ray_client_runs_only is not None:
            self.show_ray_client_runs_only = show_ray_client_runs_only
        if paging is not None:
            self.paging = paging
        if state_filter is not None:
            self.state_filter = state_filter
        if type_filter is not None:
            self.type_filter = type_filter
        if sort_by_clauses is not None:
            self.sort_by_clauses = sort_by_clauses

    @property
    def name(self):
        """Gets the name of this JobsQuery.  # noqa: E501

        Filters Jobs by name. If this field is absent, no filtering is done.  # noqa: E501

        :return: The name of this JobsQuery.  # noqa: E501
        :rtype: TextQuery
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this JobsQuery.

        Filters Jobs by name. If this field is absent, no filtering is done.  # noqa: E501

        :param name: The name of this JobsQuery.  # noqa: E501
        :type: TextQuery
        """

        self._name = name

    @property
    def runtime_environment_id(self):
        """Gets the runtime_environment_id of this JobsQuery.  # noqa: E501

        Filters Jobs by runtime enviornment id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The runtime_environment_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._runtime_environment_id

    @runtime_environment_id.setter
    def runtime_environment_id(self, runtime_environment_id):
        """Sets the runtime_environment_id of this JobsQuery.

        Filters Jobs by runtime enviornment id. If this field is absent, no filtering is done.  # noqa: E501

        :param runtime_environment_id: The runtime_environment_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._runtime_environment_id = runtime_environment_id

    @property
    def cluster_id(self):
        """Gets the cluster_id of this JobsQuery.  # noqa: E501

        Filters Jobs by cluster id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The cluster_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this JobsQuery.

        Filters Jobs by cluster id. If this field is absent, no filtering is done.  # noqa: E501

        :param cluster_id: The cluster_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def creator_id(self):
        """Gets the creator_id of this JobsQuery.  # noqa: E501

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The creator_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this JobsQuery.

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :param creator_id: The creator_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._creator_id = creator_id

    @property
    def ray_job_id(self):
        """Gets the ray_job_id of this JobsQuery.  # noqa: E501

        Filters Jobs by ray_job_id. If this field is absent, no filtering is done. Note: the ray_job_id is only unique for one cluster.  # noqa: E501

        :return: The ray_job_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._ray_job_id

    @ray_job_id.setter
    def ray_job_id(self, ray_job_id):
        """Sets the ray_job_id of this JobsQuery.

        Filters Jobs by ray_job_id. If this field is absent, no filtering is done. Note: the ray_job_id is only unique for one cluster.  # noqa: E501

        :param ray_job_id: The ray_job_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._ray_job_id = ray_job_id

    @property
    def project_id(self):
        """Gets the project_id of this JobsQuery.  # noqa: E501

        Filters Jobs by project_id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The project_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this JobsQuery.

        Filters Jobs by project_id. If this field is absent, no filtering is done.  # noqa: E501

        :param project_id: The project_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def include_child_jobs(self):
        """Gets the include_child_jobs of this JobsQuery.  # noqa: E501

        Include jobs that have parents  # noqa: E501

        :return: The include_child_jobs of this JobsQuery.  # noqa: E501
        :rtype: bool
        """
        return self._include_child_jobs

    @include_child_jobs.setter
    def include_child_jobs(self, include_child_jobs):
        """Sets the include_child_jobs of this JobsQuery.

        Include jobs that have parents  # noqa: E501

        :param include_child_jobs: The include_child_jobs of this JobsQuery.  # noqa: E501
        :type: bool
        """

        self._include_child_jobs = include_child_jobs

    @property
    def ha_job_id(self):
        """Gets the ha_job_id of this JobsQuery.  # noqa: E501

        Filter by the anyscale job  # noqa: E501

        :return: The ha_job_id of this JobsQuery.  # noqa: E501
        :rtype: str
        """
        return self._ha_job_id

    @ha_job_id.setter
    def ha_job_id(self, ha_job_id):
        """Sets the ha_job_id of this JobsQuery.

        Filter by the anyscale job  # noqa: E501

        :param ha_job_id: The ha_job_id of this JobsQuery.  # noqa: E501
        :type: str
        """

        self._ha_job_id = ha_job_id

    @property
    def show_ray_client_runs_only(self):
        """Gets the show_ray_client_runs_only of this JobsQuery.  # noqa: E501

        DEPRECATED: use type_filter. Shows only ray client runs. Orthogonal to passing ha_job_id  # noqa: E501

        :return: The show_ray_client_runs_only of this JobsQuery.  # noqa: E501
        :rtype: bool
        """
        return self._show_ray_client_runs_only

    @show_ray_client_runs_only.setter
    def show_ray_client_runs_only(self, show_ray_client_runs_only):
        """Sets the show_ray_client_runs_only of this JobsQuery.

        DEPRECATED: use type_filter. Shows only ray client runs. Orthogonal to passing ha_job_id  # noqa: E501

        :param show_ray_client_runs_only: The show_ray_client_runs_only of this JobsQuery.  # noqa: E501
        :type: bool
        """

        self._show_ray_client_runs_only = show_ray_client_runs_only

    @property
    def paging(self):
        """Gets the paging of this JobsQuery.  # noqa: E501

        Pagination information.  # noqa: E501

        :return: The paging of this JobsQuery.  # noqa: E501
        :rtype: PageQuery
        """
        return self._paging

    @paging.setter
    def paging(self, paging):
        """Sets the paging of this JobsQuery.

        Pagination information.  # noqa: E501

        :param paging: The paging of this JobsQuery.  # noqa: E501
        :type: PageQuery
        """

        self._paging = paging

    @property
    def state_filter(self):
        """Gets the state_filter of this JobsQuery.  # noqa: E501

        Filter Jobs by Job Status. If this field is an empty set, no filtering is done.  # noqa: E501

        :return: The state_filter of this JobsQuery.  # noqa: E501
        :rtype: list[BaseJobStatus]
        """
        return self._state_filter

    @state_filter.setter
    def state_filter(self, state_filter):
        """Sets the state_filter of this JobsQuery.

        Filter Jobs by Job Status. If this field is an empty set, no filtering is done.  # noqa: E501

        :param state_filter: The state_filter of this JobsQuery.  # noqa: E501
        :type: list[BaseJobStatus]
        """

        self._state_filter = state_filter

    @property
    def type_filter(self):
        """Gets the type_filter of this JobsQuery.  # noqa: E501

        Filter Jobs by their type. Their type is determined by their usage within the product e.g. Interactive sessions, job runs  # noqa: E501

        :return: The type_filter of this JobsQuery.  # noqa: E501
        :rtype: list[JobRunType]
        """
        return self._type_filter

    @type_filter.setter
    def type_filter(self, type_filter):
        """Sets the type_filter of this JobsQuery.

        Filter Jobs by their type. Their type is determined by their usage within the product e.g. Interactive sessions, job runs  # noqa: E501

        :param type_filter: The type_filter of this JobsQuery.  # noqa: E501
        :type: list[JobRunType]
        """

        self._type_filter = type_filter

    @property
    def sort_by_clauses(self):
        """Gets the sort_by_clauses of this JobsQuery.  # noqa: E501

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :return: The sort_by_clauses of this JobsQuery.  # noqa: E501
        :rtype: list[SortByClauseJobsSortField]
        """
        return self._sort_by_clauses

    @sort_by_clauses.setter
    def sort_by_clauses(self, sort_by_clauses):
        """Sets the sort_by_clauses of this JobsQuery.

        The order used to specify results. The list will be used to construct ORDER BY database queries. If not specified, the fallback order by clauses are 1. Creation time (desc) 2. Name (ascending) and 3. ID (ascending)  # noqa: E501

        :param sort_by_clauses: The sort_by_clauses of this JobsQuery.  # noqa: E501
        :type: list[SortByClauseJobsSortField]
        """

        self._sort_by_clauses = sort_by_clauses

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobsQuery):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobsQuery):
            return True

        return self.to_dict() != other.to_dict()



from unittest.mock import ANY, Mock

from anyscale.cluster_compute import (
    get_cluster_compute_from_name,
    get_default_cluster_compute,
    register_compute_template,
)
from anyscale.sdk.anyscale_client import ArchiveStatus, CreateComputeTemplate
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)


def test_get_default_cluster_compute():
    mock_api_client = Mock()
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.get_default_compute_config = Mock(
        return_value=Mock(result="mock_config_obj")
    )
    mock_default_config_obj = {
        "cloud_id": "cld_1",
        "max_workers": 10,
        "region": "us-west-2",
        "allowed_azs": ["us-west-2a", "us-west-2b"],
        "head_node_type": {
            "name": "head-1",
            "instance_type": "m5.2xlarge",
            "aws_advanced_configurations": None,
            "gcp_advanced_configurations": None,
            "resources": None,
        },
        "worker_node_types": [],
    }
    mock_anyscale_api_client.get_default_compute_config = Mock(
        return_value=Mock(result=mock_default_config_obj)
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result="mock_compute_template")
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=Mock(result=Mock(id="test_cloud_id"))
    )

    assert (
        get_default_cluster_compute(
            "mock_cloud_name", None, mock_api_client, mock_anyscale_api_client
        )
        == "mock_compute_template"
    )
    mock_anyscale_api_client.get_default_compute_config.assert_called_once_with(
        "test_cloud_id"
    )


def test_register_compute_template():
    mock_api_client = Mock()
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result=Mock(id="mock_template_id"))
    )

    assert (
        register_compute_template("mock_template_obj", mock_api_client).id
        == "mock_template_id"
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post.assert_called_once_with(
        create_compute_template=CreateComputeTemplate(
            name=ANY, config="mock_template_obj", anonymous=True,
        )
    )


def test_get_cluster_compute_from_name():
    mock_api_client = Mock()
    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(results=[Mock(id="mock_id")])
    )
    assert get_cluster_compute_from_name("mock_name", mock_api_client).id == "mock_id"

    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_once_with(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": "mock_name"},
            include_anonymous=True,
            archive_status=ArchiveStatus.ALL,
        )
    )



# type: ignore

import os
from unittest.mock import Mock, mock_open, patch

import boto3
import smart_open

import anyscale
from anyscale.utils.runtime_env import override_runtime_env_for_local_working_dir


def test_override_no_workspace():
    local_test_zip_path = "file://test.zip"
    runtime_env = {"working_dir": local_test_zip_path}
    modified_runtime_env = override_runtime_env_for_local_working_dir(runtime_env)
    assert modified_runtime_env["working_dir"] == local_test_zip_path


def test_override_upload_no_workspace():
    runtime_env = {
        "working_dir": "job-services-cuj-examples",
        "upload_path": "s3://bk-premerge-first-jawfish-artifacts/e2e_tests/job",
    }
    boto_mock = Mock()
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.multiple(
        os, makedirs=Mock()
    ), patch("builtins.open", mock_open()), patch.multiple(
        boto3, client=boto_mock
    ), patch.multiple(
        smart_open, open=Mock()
    ):
        modified_runtime_env = override_runtime_env_for_local_working_dir(runtime_env)
        assert modified_runtime_env["working_dir"].startswith(
            "s3://bk-premerge-first-jawfish-artifacts/e2e_tests/job"
        )
        assert modified_runtime_env["working_dir"].endswith(".zip")


def test_override_upload_with_workspace():
    runtime_env = {
        "working_dir": "job-services-cuj-examples",
        "upload_path": "s3://bk-premerge-first-jawfish-artifacts/e2e_tests/job",
    }
    boto_mock = Mock()
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.dict(
        os.environ,
        {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "Test_workspace_id"},
        clear=True,
    ), patch.multiple(os, makedirs=Mock()), patch(
        "builtins.open", mock_open()
    ), patch.multiple(
        boto3, client=boto_mock
    ), patch.multiple(
        smart_open, open=Mock()
    ):
        modified_runtime_env = override_runtime_env_for_local_working_dir(runtime_env)
        assert modified_runtime_env["working_dir"].startswith("file:///efs/jobs")
        assert modified_runtime_env["working_dir"].endswith(".zip")


def test_runtime_env_override_to_efs_workspace():
    runtime_env = {"working_dir": "file://test.zip"}
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.dict(
        os.environ,
        {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "Test_workspace_id"},
        clear=True,
    ), patch.multiple(os, makedirs=Mock()), patch(
        "urllib.request", urlretrieve=Mock()
    ), patch(
        "builtins.open", mock_open()
    ):
        modified_runtime_env = override_runtime_env_for_local_working_dir(runtime_env)
        assert modified_runtime_env["working_dir"].startswith("file:///efs")


def test_local_runtime_env_override_to_efs_job():
    runtime_env = {"working_dir": "file://test.zip"}
    test_job_id = "Test_job_id"
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_INITIAL_JOB_ID": test_job_id}, clear=True,
    ), patch.multiple(os, makedirs=Mock()), patch(
        "urllib.request", urlretrieve=Mock()
    ), patch(
        "builtins.open", mock_open()
    ):
        modified_runtime_env = anyscale.snapshot_util.env_hook(runtime_env)
        assert (
            modified_runtime_env["working_dir"]
            == f"/efs/jobs/{test_job_id}/working_dir.zip"
        )


def test_s3_runtime_env_override_to_efs_job():
    runtime_env = {
        "working_dir": "s3://bk-premerge-first-jawfish-artifacts/e2e_tests/job/_anyscale_pkg_00216a215e75dff900a133c9ac9c764a.zip"
    }
    test_job_id = "Test_job_id"
    boto_mock = Mock()
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_INITIAL_JOB_ID": test_job_id}, clear=True,
    ), patch.multiple(os, makedirs=Mock()), patch(
        "builtins.open", mock_open()
    ), patch.multiple(
        boto3, client=boto_mock
    ):
        modified_runtime_env = anyscale.snapshot_util.env_hook(runtime_env)
        boto_mock.assert_called_once_with("s3")
        assert (
            modified_runtime_env["working_dir"]
            == f"/efs/jobs/{test_job_id}/working_dir.zip"
        )


def test_https_runtime_env_override_to_efs_job():
    runtime_env = {
        "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip"
    }
    test_job_id = "Test_job_id"
    urlretrieve_mock = Mock()
    with patch.multiple(os.path, ismount=Mock(return_value=True)), patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_INITIAL_JOB_ID": test_job_id}, clear=True,
    ), patch.multiple(os, makedirs=Mock()), patch(
        "builtins.open", mock_open()
    ), patch.multiple(
        "urllib.request", urlretrieve=urlretrieve_mock
    ):
        modified_runtime_env = anyscale.snapshot_util.env_hook(runtime_env)
        urlretrieve_mock.assert_called_once()
        assert (
            modified_runtime_env["working_dir"]
            == f"/efs/jobs/{test_job_id}/working_dir.zip"
        )
